(*
  This file is distributed under the terms of the
  GNU Lesser General Public License version 2.1.
*)

(*
  A finite sets library for Coq.
  Copyright (C) 2006 Brian Aydemir and Aaron Bohannon

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*)

(*
  This file is based on the following files, originally distributed
  with Coq 8.1beta:

    (* Id: OrderedType.v 8834 2006-05-20 00:41:35Z letouzey *)
    (* Id: FSetInterface.v 8820 2006-05-15 11:44:05Z letouzey *)
    (* Id: FSetList.v 8834 2006-05-20 00:41:35Z letouzey *)

  The modifications here are by:
    Brian Aydemir ( http://www.cis.upenn.edu/~baydemir/ )

  The most recent modifcations were made on:
  $Date: 2006-08-24 14:38:25 -0400 (Thu, 24 Aug 2006) $
  $Id: ExtFsetImpl.v 766 2006-08-24 18:38:25Z baydemir $
*)

Require Import Bool.
Require Import List.
Require Import OrderedType.
Require Import SetoidList.

Require Import Axioms.
Require Export ExtFset.

Set Implicit Arguments.
Unset Strict Implicit.

(**
  The main purpose of this file is to provide an implementation of the
  following module signature.  We use modules here to purposefully keep the
  implementation opaque.
*)
Module Type ExtFsetSig.
Parameter mkOrderedListExtFset
     : forall (elt : Set) (lt : elt -> elt -> Prop),
       (forall x y z : elt, lt x y -> lt y z -> lt x z) ->
       (forall x y : elt, lt x y -> x <> y) ->
       (forall x y : elt, Compare lt (eq (A:=elt)) x y) -> ExtFset elt.
End ExtFsetSig.

Module ExtFsetImplementation : ExtFsetSig.

(**
  Hide the implementation for documentation purposes.  Note that we need proof
  irrelevance here.
*)
(* begin hide *)

Section Implementation.

Variable elt : Set.

Definition eq_refl := (@refl_equal elt).
Definition eq_sym := (@sym_eq elt).
Definition eq_trans := (@trans_eq elt).

(* ********************************************************************** *)
(** ** From FsetInterface.v *)

Definition compat_bool (A:Set)(eqA: A->A->Prop)(f: A-> bool) :=
  forall x y : A, eqA x y -> f x = f y.

Definition compat_P (A:Set)(eqA: A->A->Prop)(P : A -> Prop) :=
  forall x y : A, eqA x y -> P x -> P y.

Hint Unfold compat_bool compat_P.

(* ********************************************************************** *)
(** ** From OrderedType.v *)

  Variable lt : elt -> elt -> Prop.

  Hypothesis lt_trans : forall x y z : elt, lt x y -> lt y z -> lt x z.
  Hypothesis lt_not_eq : forall x y : elt, lt x y -> ~ eq x y.

  Variable compare : forall x y : elt, Compare lt (@eq elt) x y.

  Hint Immediate eq_sym.
  Hint Resolve eq_refl eq_trans lt_not_eq lt_trans.

  Lemma lt_antirefl : forall x, ~ lt x x.
  Proof.
   intros; intro; absurd (eq x x); auto.
  Qed.

  Lemma lt_eq : forall x y z, lt x y -> eq y z -> lt x z.
  Proof.
   intros; destruct (compare x z); auto.
   elim (lt_not_eq H); apply eq_trans with z; auto.
   elim (lt_not_eq (lt_trans l H)); auto.
  Qed.

  Lemma eq_lt : forall x y z, eq x y -> lt y z -> lt x z.
  Proof.
   intros; destruct (compare x z); auto.
   elim (lt_not_eq H0); apply eq_trans with x; auto.
   elim (lt_not_eq (lt_trans H0 l)); auto.
  Qed.

  Lemma le_eq : forall x y z, ~lt x y -> eq y z -> ~lt x z.
  Proof.
   intros; intro; destruct H; apply lt_eq with z; auto.
  Qed.

  Lemma eq_le : forall x y z, eq x y -> ~lt y z -> ~lt x z.
  Proof.
   intros; intro; destruct H0; apply eq_lt with x; auto.
  Qed.

  Lemma neq_eq : forall x y z : elt, ~eq x y -> eq y z -> ~eq x z.
  Proof.
   intros; intro; destruct H; apply eq_trans with z; auto.
  Qed.

  Lemma eq_neq : forall x y z : elt, eq x y -> ~eq y z -> ~eq x z.
  Proof.
   intros; intro; destruct H0; apply eq_trans with x; auto.
  Qed.

  Hint Immediate eq_lt lt_eq le_eq eq_le neq_eq eq_neq.

  Lemma le_lt_trans : forall x y z, ~lt y x -> lt y z -> lt x z.
  Proof.
   intros; destruct (compare y x); auto.
   elim (H l).
   apply eq_lt with y; auto.
   apply lt_trans with y; auto.
  Qed.

  Lemma lt_le_trans : forall x y z, lt x y -> ~lt z y -> lt x z.
  Proof.
   intros; destruct (compare z y); auto.
   elim (H0 l).
   apply lt_eq with y; auto.
   apply lt_trans with y; auto.
  Qed.

  Lemma le_neq : forall x y, ~lt x y -> ~eq x y -> lt y x.
  Proof.
   intros; destruct (compare x y); intuition.
  Qed.

  Lemma neq_sym : forall x y : elt, ~eq x y -> ~eq y x.
  Proof.
   intuition.
  Qed.

Ltac abstraction := match goal with
 (* First, some obvious simplifications *)
 | H : False |- _ => elim H
 | H : lt ?x ?x |- _ => elim (lt_antirefl H)
 | H : ~eq ?x ?x |- _ => elim (H (eq_refl x))
 | H : eq ?x ?x |- _ => clear H; abstraction
 | H : ~lt ?x ?x |- _ => clear H; abstraction
 | |- eq ?x ?x => exact (eq_refl x)
 | |- lt ?x ?x => elimtype False; abstraction
 | |- ~ _ => intro; abstraction
 | H1: ~lt ?x ?y, H2: ~eq ?x ?y |- _ =>
     generalize (le_neq H1 H2); clear H1 H2; intro; abstraction
 | H1: ~lt ?x ?y, H2: ~eq ?y ?x |- _ =>
     generalize (le_neq H1 (neq_sym H2)); clear H1 H2; intro; abstraction
 (* Then, we generalize all interesting facts *)
 | H : lt ?x ?y |- _ => revert H; abstraction
 | H : ~lt ?x ?y |- _ => revert H; abstraction
 | H : ~eq ?x ?y |- _ =>  revert H; abstraction
 | H : eq ?x ?y |- _ =>  revert H; abstraction
 | _ => idtac
end.

Ltac do_eq a b EQ := match goal with
 | |- lt ?x ?y -> _ => let H := fresh "H" in
     (intro H;
      (generalize (eq_lt (eq_sym EQ) H); clear H; intro H) ||
      (generalize (lt_eq H EQ); clear H; intro H) ||
      idtac);
      do_eq a b EQ
 | |- ~lt ?x ?y -> _ => let H := fresh "H" in
     (intro H;
      (generalize (eq_le (eq_sym EQ) H); clear H; intro H) ||
      (generalize (le_eq H EQ); clear H; intro H) ||
      idtac);
      do_eq a b EQ
 | |- eq ?x ?y -> _ => let H := fresh "H" in
     (intro H;
      (generalize (eq_trans (eq_sym EQ) H); clear H; intro H) ||
      (generalize (eq_trans H EQ); clear H; intro H) ||
      idtac);
      do_eq a b EQ
 | |- ~eq ?x ?y -> _ => let H := fresh "H" in
     (intro H;
      (generalize (eq_neq (eq_sym EQ) H); clear H; intro H) ||
      (generalize (neq_eq H EQ); clear H; intro H) ||
      idtac);
      do_eq a b EQ
 | |- lt a ?y => apply eq_lt with b; [exact EQ|]
 | |- lt ?y a => apply lt_eq with b; [|exact (eq_sym EQ)]
 | |- eq a ?y => apply eq_trans with b; [exact EQ|]
 | |- eq ?y a => apply eq_trans with b; [|exact (eq_sym EQ)]
 | _ => idtac
 end.

Ltac propagate_eq := abstraction; match goal with
 (* the abstraction tactic leaves equality facts in head position...*)
 | |- eq ?a ?b -> _ =>
     let EQ := fresh "EQ" in (intro EQ; do_eq a b EQ; clear EQ);
     propagate_eq
 | _ => idtac
end.

Ltac do_lt x y LT := match goal with
 (* LT *)
 | |- lt x y -> _ => intros _; do_lt x y LT
 | |- lt y ?z -> _ => let H := fresh "H" in
     (intro H; generalize (lt_trans LT H); intro); do_lt x y LT
 | |- lt ?z x -> _ => let H := fresh "H" in
     (intro H; generalize (lt_trans H LT); intro); do_lt x y LT
 | |- lt _ _ -> _ => intro; do_lt x y LT
 (* Ge *)
 | |- ~lt y x -> _ => intros _; do_lt x y LT
 | |- ~lt x ?z -> _ => let H := fresh "H" in
     (intro H; generalize (le_lt_trans H LT); intro); do_lt x y LT
 | |- ~lt ?z y -> _ => let H := fresh "H" in
     (intro H; generalize (lt_le_trans LT H); intro); do_lt x y LT
 | |- ~lt _ _ -> _ => intro; do_lt x y LT
 | _ => idtac
 end.

Definition hide_lt := lt.

Ltac propagate_lt := abstraction; match goal with
 (* when no [=] remains, the abstraction tactic leaves [<] facts first. *)
 | |- lt ?x ?y -> _ =>
     let LT := fresh "LT" in (intro LT; do_lt x y LT;
     change (hide_lt x y) in LT);
     propagate_lt
 | _ => unfold hide_lt in *
end.

Ltac order :=
 intros;
 propagate_eq;
 propagate_lt;
 auto;
 propagate_lt;
 eauto.

Ltac false_order := elimtype False; order.

  Lemma gt_not_eq : forall x y, lt y x -> ~ eq x y.
  Proof.
    order.
  Qed.

  Lemma eq_not_lt : forall x y : elt, eq x y -> ~ lt x y.
  Proof.
    order.
  Qed.

  Hint Resolve gt_not_eq eq_not_lt.

  Lemma eq_not_gt : forall x y : elt, eq x y -> ~ lt y x.
  Proof.
   order.
  Qed.

  Lemma lt_not_gt : forall x y : elt, lt x y -> ~ lt y x.
  Proof.
   order.
  Qed.

  Hint Resolve eq_not_gt lt_antirefl lt_not_gt.

  Lemma elim_compare_eq :
   forall x y : elt,
   eq x y -> exists H : eq x y, compare x y = EQ _ H.
  Proof.
   intros; case (compare x y); intros H'; try solve [false_order].
   exists H'; auto.
  Qed.

  Lemma elim_compare_lt :
   forall x y : elt,
   lt x y -> exists H : lt x y, compare x y = LT _ H.
  Proof.
   intros; case (compare x y); intros H'; try solve [false_order].
   exists H'; auto.
  Qed.

  Lemma elim_compare_gt :
   forall x y : elt,
   lt y x -> exists H : lt y x, compare x y = GT _ H.
  Proof.
   intros; case (compare x y); intros H'; try solve [false_order].
   exists H'; auto.
  Qed.

  Ltac elim_comp :=
    match goal with
      | |- ?e => match e with
           | context ctx [ compare ?a ?b ] =>
                let H := fresh in
                (destruct (compare a b) as [H|H|H];
                 try solve [ intros; false_order])
         end
    end.

  Ltac elim_comp_eq x y :=
    elim (elim_compare_eq (x:=x) (y:=y));
     [ intros _1 _2; rewrite _2; clear _1 _2 | auto ].

  Ltac elim_comp_lt x y :=
    elim (elim_compare_lt (x:=x) (y:=y));
     [ intros _1 _2; rewrite _2; clear _1 _2 | auto ].

  Ltac elim_comp_gt x y :=
    elim (elim_compare_gt (x:=x) (y:=y));
     [ intros _1 _2; rewrite _2; clear _1 _2 | auto ].

  Lemma eq_dec : forall x y : elt, {eq x y} + {~ eq x y}.
  Proof.
   intros; elim (compare x y); [ right | left | right ]; auto.
  Qed.

  Lemma lt_dec : forall x y : elt, {lt x y} + {~ lt x y}.
  Proof.
   intros; elim (compare x y); [ left | right | right ]; auto.
  Qed.

  Definition eqb x y : bool := if eq_dec x y then true else false.

  Lemma eqb_alt :
    forall x y, eqb x y = match compare x y with EQ _ => true | _ => false end.
  Proof.
  unfold eqb; intros; destruct (eq_dec x y); elim_comp; auto.
  Qed.

(* Specialization of resuts about lists modulo. *)

Section ForNotations.

Notation In:=(InA (@eq elt)).
Notation Inf:=(lelistA lt).
Notation Sort:=(sort lt).
Notation NoDup:=(NoDupA (@eq elt)).

Lemma In_eq : forall l x y, eq x y -> In x l -> In y l.
Proof. exact (InA_eqA eq_sym eq_trans). Qed.

Lemma ListIn_In : forall l x, List.In x l -> In x l.
Proof. exact (In_InA eq_refl). Qed.

Lemma Inf_lt : forall l x y, lt x y -> Inf y l -> Inf x l.
Proof. exact (InfA_ltA lt_trans). Qed.

Lemma Inf_eq : forall l x y, eq x y -> Inf y l -> Inf x l.
Proof. exact (InfA_eqA eq_lt). Qed.

Lemma Sort_Inf_In : forall l x a, Sort l -> Inf a l -> In x l -> lt a x.
Proof. exact (SortA_InfA_InA eq_refl eq_sym lt_trans lt_eq eq_lt). Qed.

Lemma ListIn_Inf : forall l x, (forall y, List.In y l -> lt x y) -> Inf x l.
Proof. exact (@In_InfA elt lt). Qed.

Lemma In_Inf : forall l x, (forall y, In y l -> lt x y) -> Inf x l.
Proof. exact (InA_InfA eq_refl (ltA:=lt)). Qed.

Lemma Inf_alt :
 forall l x, Sort l -> (Inf x l <-> (forall y, In y l -> lt x y)).
Proof. exact (InfA_alt eq_refl eq_sym lt_trans lt_eq eq_lt). Qed.

Lemma Sort_NoDup : forall l, Sort l -> NoDup l.
Proof. exact (SortA_NoDupA eq_refl eq_sym lt_trans lt_not_eq lt_eq eq_lt) . Qed.

End ForNotations.

Hint Resolve ListIn_In Sort_NoDup Inf_lt.
Hint Immediate In_eq Inf_lt.

(* ********************************************************************** *)
(** ** From FSetList.v *)

  Definition t := list elt.

  Definition empty : t := nil.

  Fixpoint add (x : elt) (s : t) {struct s} : t :=
    match s with
    | nil => x :: nil
    | y :: l =>
        match compare x y with
        | LT _ => x :: s
        | EQ _ => s
        | GT _ => y :: add x l
        end
    end.

  Fixpoint union (s : t) : t -> t :=
    match s with
    | nil => fun s' => s'
    | x :: l =>
        (fix union_aux (s' : t) : t :=
           match s' with
           | nil => s
           | x' :: l' =>
               match compare x x' with
               | LT _ => x :: union l s'
               | EQ _ => x :: union l l'
               | GT _ => x' :: union_aux l'
               end
           end)
    end.

  Fixpoint inter (s : t) : t -> t :=
    match s with
    | nil => fun _ => nil
    | x :: l =>
        (fix inter_aux (s' : t) : t :=
           match s' with
           | nil => nil
           | x' :: l' =>
               match compare x x' with
               | LT _ => inter l s'
               | EQ _ => x :: inter l l'
               | GT _ => inter_aux l'
               end
           end)
    end.

  Fixpoint diff (s : t) : t -> t :=
    match s with
    | nil => fun _ => nil
    | x :: l =>
        (fix diff_aux (s' : t) : t :=
           match s' with
           | nil => s
           | x' :: l' =>
               match compare x x' with
               | LT _ => x :: diff l s'
               | EQ _ => diff l l'
               | GT _ => diff_aux l'
               end
           end)
    end.

  Fixpoint equal (s : t) : t -> bool :=
    fun s' : t =>
    match s, s' with
    | nil, nil => true
    | x :: l, x' :: l' =>
        match compare x x' with
        | EQ _ => equal l l'
        | _ => false
        end
    | _, _ => false
    end.

  Definition elements (x : t) : list elt := x.

  Section ForNotations_in_FSetList.

  Notation Sort := (sort lt).
  Notation Inf := (lelistA lt).
  Notation In := (InA (@eq elt)).

  Definition Equal s s' := forall a : elt, In a s <-> In a s'.

  Lemma equal_1 :
   forall (s s' : t) (Hs : Sort s) (Hs' : Sort s'),
   Equal s s' -> equal s s' = true.
  Proof.
  simple induction s; unfold Equal.
  intro s'; case s'; auto.
  simpl; intuition.
  elim (H e); intros; assert (A : In e nil); auto; inversion A.
  intros x l Hrec s'.
  case s'.
  intros; elim (H x); intros; assert (A : In x nil); auto; inversion A.
  intros x' l' Hs Hs'; inversion Hs; inversion Hs'; subst.
  simpl; case (compare x); intros; auto.

  elim (H x); intros.
  assert (A : In x (x' :: l')); auto; inversion_clear A.
  order.
  generalize (Sort_Inf_In H5 H6 H4); order.

  apply Hrec; intuition; elim (H a); intros.
  assert (A : In a (x' :: l')); auto; inversion_clear A; auto.
  generalize (Sort_Inf_In H1 H2 H0); order.
  assert (A : In a (x :: l)); auto; inversion_clear A; auto.
  generalize (Sort_Inf_In H5 H6 H0); order.

  elim (H x'); intros.
  assert (A : In x' (x :: l)); auto; inversion_clear A.
  order.
  generalize (Sort_Inf_In H1 H2 H4); order.
  Qed.

  Lemma add_Inf :
   forall (s : t) (x a : elt), Inf a s -> lt a x -> Inf a (add x s).
  Proof.
  simple induction s.
  simpl; intuition.
  simpl; intros; case (compare x a); intuition; inversion H0;
   intuition.
  Qed.
  Hint Resolve add_Inf.

  Lemma add_sort : forall (s : t) (Hs : Sort s) (x : elt), Sort (add x s).
  Proof.
  simple induction s.
  simpl; intuition.
  simpl; intros; case (compare x a); intuition; inversion_clear Hs;
   auto.
  Qed.

  Lemma add_1 :
   forall (s : t) (Hs : Sort s) (x y : elt), eq x y -> In y (add x s).
  Proof.
  simple induction s.
  simpl; intuition.
  simpl; intros; case (compare x a); inversion_clear Hs; auto.
  constructor; apply eq_trans with x; auto.
  Qed.

  Lemma add_2 :
   forall (s : t) (Hs : Sort s) (x y : elt), In y s -> In y (add x s).
  Proof.
  simple induction s.
  simpl; intuition.
  simpl; intros; case (compare x a); intuition.
  inversion_clear Hs; inversion_clear H0; auto.
  Qed.

  Lemma add_3 :
   forall (s : t) (Hs : Sort s) (x y : elt),
   ~ eq x y -> In y (add x s) -> In y s.
  Proof.
  simple induction s.
  simpl; inversion_clear 3; auto; order.
  simpl; intros a l Hrec Hs x y; case (compare x a); intros;
   inversion_clear H0; inversion_clear Hs; auto.
  order.
  constructor 2; apply Hrec with x; auto.
  Qed.

  Ltac DoubleInd :=
    simple induction s;
     [ simpl; auto; try solve [ intros; inversion H ]
     | intros x l Hrec; simple induction s';
        [ simpl; auto; try solve [ intros; inversion H ]
        | intros x' l' Hrec' Hs Hs'; inversion Hs; inversion Hs'; subst;
           simpl ] ].

  Lemma union_Inf :
   forall (s s' : t) (Hs : Sort s) (Hs' : Sort s') (a : elt),
   Inf a s -> Inf a s' -> Inf a (union s s').
  Proof.
  DoubleInd.
  intros i His His'; inversion_clear His; inversion_clear His'.
  case (compare x x'); auto.
  Qed.
  Hint Resolve union_Inf.

  Lemma union_sort :
   forall (s s' : t) (Hs : Sort s) (Hs' : Sort s'), Sort (union s s').
  Proof.
  DoubleInd; case (compare x x'); intuition; constructor; auto.
  apply Inf_eq with x'; trivial; apply union_Inf; trivial; apply Inf_eq with x; auto.
  change (Inf x' (union (x :: l) l')); auto.
  Qed.

  Lemma union_1 :
   forall (s s' : t) (Hs : Sort s) (Hs' : Sort s') (x : elt),
   In x (union s s') -> In x s \/ In x s'.
  Proof.
  DoubleInd; case (compare x x'); intuition; inversion_clear H; intuition.
  elim (Hrec (x' :: l') H1 Hs' x0); intuition.
  elim (Hrec l' H1 H5 x0); intuition.
  elim (H0 x0); intuition.
  Qed.

  Lemma union_2 :
   forall (s s' : t) (Hs : Sort s) (Hs' : Sort s') (x : elt),
   In x s -> In x (union s s').
  Proof.
  DoubleInd.
  intros i Hi; case (compare x x'); intuition; inversion_clear Hi; auto.
  Qed.

  Lemma union_3 :
   forall (s s' : t) (Hs : Sort s) (Hs' : Sort s') (x : elt),
   In x s' -> In x (union s s').
  Proof.
  DoubleInd.
  intros i Hi; case (compare x x'); inversion_clear Hi; intuition.
  constructor; apply eq_trans with x'; auto.
  Qed.

  Lemma inter_Inf :
   forall (s s' : t) (Hs : Sort s) (Hs' : Sort s') (a : elt),
   Inf a s -> Inf a s' -> Inf a (inter s s').
  Proof.
  DoubleInd.
  intros i His His'; inversion His; inversion His'; subst.
  case (compare x x'); intuition.
  apply Inf_lt with x; auto.
  apply H3; auto.
  apply Inf_lt with x'; auto.
  Qed.
  Hint Resolve inter_Inf.

  Lemma inter_sort :
   forall (s s' : t) (Hs : Sort s) (Hs' : Sort s'), Sort (inter s s').
  Proof.
  DoubleInd; case (compare x x'); auto.
  constructor; auto.
  apply Inf_eq with x'; trivial; apply inter_Inf; trivial; apply Inf_eq with x; auto.
  Qed.

  Lemma inter_1 :
   forall (s s' : t) (Hs : Sort s) (Hs' : Sort s') (x : elt),
   In x (inter s s') -> In x s.
  Proof.
  DoubleInd; case (compare x x'); intuition.
  constructor 2; apply Hrec with (x'::l'); auto.
  inversion_clear H; auto.
  constructor 2; apply Hrec with l'; auto.
  Qed.

  Lemma inter_2 :
   forall (s s' : t) (Hs : Sort s) (Hs' : Sort s') (x : elt),
   In x (inter s s') -> In x s'.
  Proof.
  DoubleInd; case (compare x x'); intuition; inversion_clear H.
  constructor 1; apply eq_trans with x; auto.
  constructor 2; auto.
  Qed.

  Lemma inter_3 :
   forall (s s' : t) (Hs : Sort s) (Hs' : Sort s') (x : elt),
   In x s -> In x s' -> In x (inter s s').
  Proof.
  DoubleInd.
  intros i His His'; elim (compare x x'); intuition.

  inversion_clear His; auto.
  generalize (Sort_Inf_In Hs' (cons_leA _ _ _ _ l0) His'); order.

  inversion_clear His; auto; inversion_clear His'; auto.
  constructor; apply eq_trans with x'; auto.

  change (In i (inter (x :: l) l')).
  inversion_clear His'; auto.
  generalize (Sort_Inf_In Hs (cons_leA _ _ _ _ l0) His); order.
  Qed.

  Lemma diff_Inf :
   forall (s s' : t) (Hs : Sort s) (Hs' : Sort s') (a : elt),
   Inf a s -> Inf a s' -> Inf a (diff s s').
  Proof.
  DoubleInd.
  intros i His His'; inversion His; inversion His'.
  case (compare x x'); intuition.
  apply Hrec; trivial.
  apply Inf_lt with x; auto.
  apply Inf_lt with x'; auto.
  apply H10; trivial.
  apply Inf_lt with x'; auto.
  Qed.
  Hint Resolve diff_Inf.

  Lemma diff_sort :
   forall (s s' : t) (Hs : Sort s) (Hs' : Sort s'), Sort (diff s s').
  Proof.
  DoubleInd; case (compare x x'); auto.
  Qed.

  Lemma diff_1 :
   forall (s s' : t) (Hs : Sort s) (Hs' : Sort s') (x : elt),
   In x (diff s s') -> In x s.
  Proof.
  DoubleInd; case (compare x x'); intuition.
  inversion_clear H; auto.
  constructor 2; apply Hrec with (x'::l'); auto.
  constructor 2; apply Hrec with l'; auto.
  Qed.

  Lemma diff_2 :
   forall (s s' : t) (Hs : Sort s) (Hs' : Sort s') (x : elt),
   In x (diff s s') -> ~ In x s'.
  Proof.
  DoubleInd.
  intros; intro Abs; inversion Abs.
  case (compare x x'); intuition.

  inversion_clear H.
  generalize (Sort_Inf_In Hs' (cons_leA _ _ _ _ l0) H3); order.
  apply Hrec with (x'::l') x0; auto.

  inversion_clear H3.
  generalize (Sort_Inf_In H1 H2 (diff_1 H1 H5 H)); order.
  apply Hrec with l' x0; auto.

  inversion_clear H3.
  generalize (Sort_Inf_In Hs (cons_leA _ _ _ _ l0) (diff_1 Hs H5 H)); order.
  apply H0 with x0; auto.
  Qed.

  Lemma diff_3 :
   forall (s s' : t) (Hs : Sort s) (Hs' : Sort s') (x : elt),
   In x s -> ~ In x s' -> In x (diff s s').
  Proof.
  DoubleInd.
  intros i His His'; elim (compare x x'); intuition; inversion_clear His; auto.
  elim His'; constructor; apply eq_trans with x; auto.
  Qed.

  Lemma empty_sort : Sort empty.
  Proof.
  unfold empty; constructor.
  Qed.

(* ********************************************************************** *)
(** ** Our additional code *)

  Lemma add_3' :
   forall (s : t) (Hs : Sort s) (x y : elt),
   In y (add x s) -> eq y x \/ In y s.
  Proof.
   intros s Hs x y H.
   case (eq_dec x y); eauto using add_3.
  Qed.

  Lemma In_equiv :
    forall x s, In x s <-> List.In x s.
  Proof.
    induction s; split; intros H.
    inversion H.
    inversion H.
    inversion H; try subst.
      simpl; auto.
      simpl; right.
      rewrite IHs in H1; assumption.
    simpl in H; inversion H.
      subst; auto.
      rewrite <- IHs in H0; auto.
  Qed.

  Lemma equal_extensional :
    forall s s', Sort s -> Sort s' -> equal s s' = true -> s = s'.
  Proof.
    induction s; induction s'; intros H J; simpl; auto.
    intros; discriminate.
    intros; discriminate.
    case (compare a a0); try (intros; discriminate).
    intros K L; subst.
    inversion H; inversion J; subst.
    rewrite (IHs s' H2 H6 L); auto.
  Qed.

  Fixpoint map (f : elt -> elt) (s : t) {struct s} : t :=
    match s with
      | nil => nil
      | h :: t => add (f h) (map f t)
    end.

  Lemma map_sort :
    forall (s : t) (Hs : Sort s) (f : elt -> elt), Sort (map f s).
  Proof.
    induction s; auto.
    intros H f; inversion H; subst; simpl.
    auto using add_sort.
  Qed.
  Hint Resolve map_sort.

  Lemma map_1 :
    forall s f x, Sort s -> In x s -> In (f x) (map f s).
  Proof.
    induction s; intros f x Hs.
    intros H; inversion H.
    intros H; inversion H; inversion Hs; subst.
    simpl; auto using add_1.
    simpl; auto using add_2.
  Qed.

  Lemma map_2 :
    forall s f y, Sort s -> In y (map f s) -> exists x, In x s /\ y = f x.
  Proof.
    induction s; intros f y H J.
    inversion J.
    simpl in J.
    inversion H; subst.
    assert (K := add_3' (map_sort H2 f) J).
    inversion K as [L | L].
    exists a; auto.
    case (IHs f y H2 L).
    intros x [M N].
    exists x; auto.
  Qed.

  End ForNotations_in_FSetList.

Record slist : Set := mkSlist { this : list elt; sorted : sort lt this }.

Definition mkOrderedListExtFset : ExtFset elt.
  refine (mkExtFset elt slist

    (* Membership tests. *)
    (fun x E => InA (@eq elt) x E.(this)) _

    (* Equality. *)
    eq_dec _

    (* Empty set. *)
    (mkSlist empty_sort) _

    (* Adding an element to a set. *)
    (fun x E => mkSlist (add_sort (sorted E) x)) _ _ _

    (* Union. *)
    (fun E F => mkSlist (union_sort (sorted E) (sorted F))) _ _ _

    (* Intersection. *)
    (fun E F => mkSlist (inter_sort (sorted E) (sorted F))) _ _ _

    (* Difference. *)
    (fun E F => mkSlist (diff_sort (sorted E) (sorted F))) _ _ _

    (* Map. *)
    (fun f E => mkSlist (map_sort (sorted E) f)) _ _

    (* Elements. *)
    (fun E => elements E.(this)) _
  ).
  intros x E; apply InA_dec.
    apply eq_dec.
  intros E F H; induction E; induction F; simpl in *.
    cut (this0 = this1).
    intros J.
    revert sorted0.
    dependent rewrite J.
    intros sorted0.
    rewrite (Proof_irrelevance _ sorted0 sorted1); reflexivity.
    apply equal_extensional; auto.
    apply equal_1; auto.
  intros x H; inversion H.
  intros x y E H; induction E; simpl.
    apply add_1; auto.
  intros x y E H; induction E; simpl.
    apply add_2; auto.
  intros x y E H; induction E; simpl.
    apply add_3'; auto.
  intros x E F H; induction E; induction F; simpl.
    apply union_2; auto.
  intros x E F H; induction E; induction F; simpl.
    apply union_3; auto.
  intros x E F H; induction E; induction F; simpl.
    apply union_1; auto.
  intros x E F H; induction E; induction F; simpl.
    apply inter_3; auto.
  intros x E F H; induction E; induction F; simpl in *.
    apply inter_1 with (s' := this1); auto.
  intros x E F H; induction E; induction F; simpl in *.
    apply inter_2 with (s := this0); auto.
  intros x E F H; induction E; induction F; simpl.
    apply diff_3; auto.
  intros x E F H; induction E; induction F; simpl.
    apply diff_1 with (s' := this1); auto.
  intros x E F H; induction E; induction F; simpl.
    apply diff_2 with (s := this0); auto.
  intros x E f H; induction E; simpl.
    apply map_1; auto.
  intros x E f H; induction E; simpl in *.
    apply map_2; auto.
  intros x E; induction E; simpl.
    rewrite In_equiv; split; auto.
Qed.

End Implementation.

(* end hide *)

End ExtFsetImplementation.
Export ExtFsetImplementation.
Implicit Arguments mkOrderedListExtFset [elt].
