Require Import Atoms.
Require Import Axioms.


(**
  The main purpose of this file is to provide an implementation of the
  following module signature.  We use modules here to purposefully keep the
  implementation opaque.

  We then prove a variety of theorems from the signature (see below).
*)
Module Type PermutationsAbstraction.

Section PermutationDefinitions.
  (** [permt A] is the type of permutations of atoms of type [A]. *)
  Parameter permt : AtomT -> Set.

  Variable A : AtomT.

  (** Apply a permutation to an atom. *)
  Parameter perma : permt A -> A -> A.

  (** Identity permutation. *)
  Parameter pid : permt A.

  (** Inverse of a permutation. *)
  Parameter pinv : permt A -> permt A.

  (** Compose two permutations. *)
  Parameter pcompose : permt A -> permt A -> permt A.

  (** Extend a permutation by a swap. *)
  Parameter pextend : (A * A) -> permt A -> permt A.

End PermutationDefinitions.

Implicit Arguments perma [A].
Implicit Arguments pid [A].
Implicit Arguments pinv [A].
Implicit Arguments pcompose [A].
Implicit Arguments pextend [A].


Notation "[]" := (pid) : perms_scope.

Arguments Scope perma [_ perms_scope _ _].
Arguments Scope pinv [_ perms_scope].
Arguments Scope pcompose [_ perms_scope perms_scope].
Arguments Scope pextend [_ _ perms_scope].

Notation "p @ a" :=
  (perma p a)
  (at level 65, right associativity) : perms_scope.

Notation "- p" :=
  (pinv p)
  (at level 35, right associativity) : perms_scope.

Notation "p :++ q" :=
  (pcompose p q)
  (at level 60, right associativity) : perms_scope.

Notation "x :+ p" :=
  (pextend x p)
  (at level 60, right associativity) : perms_scope.

Notation "[ x ; .. ; y ]" := (pextend x .. (pextend y pid) ..) : perms_scope.

Bind Scope perms_scope with permt.
Open Scope perms_scope.


(**
  Basic properties of permutations.
*)
Section PermutationBaseProperties.
  Variable A : AtomT.

  Axiom perma_id :
    forall a : A, [] @ a = a.

  Axiom perma_inverse :
    forall (p : permt A) (a : A), (- p) @ p @ a = a.

  Axiom perma_compose :
    forall (p q : permt A) (a : A), (p :++ q) @ a = p @ q @ a.

  Axiom perma_eq :
    forall (p q : permt A), (forall a, p @ a = q @ a) -> p = q.

  Axiom perma_extend :
    forall (p : permt A) (ab : A * A) (c : A),
    (ab :+ p) @ c = swapa A ab (p @ c).

End PermutationBaseProperties.

Hint Resolve perma_id perma_inverse perma_compose perma_extend : nominal.
Hint Rewrite perma_id perma_inverse perma_compose perma_extend : nominal.

End PermutationsAbstraction.


(* ********************************************************************** *)
(** ** Implementation of atom permutations *)

Module PermutationsImplementation : PermutationsAbstraction.

(**
  We hide the implementation for documentation purposes.  Note that we need
  extensional equality on functions and proof irrelevance here.
*)
(* begin hide *)

(**
  Permutations are essentially bijective functions on atoms paired with
  their inverse.  Most of the arguments to [Perm_intro] ensure that the
  given pair of functions have the required properties.  These additional
  arguments push the work of implementing permutations into the defintions
  of the operations on permutations.
*)
Inductive Perm (A : AtomT) : Set :=
  | Perm_intro :
      forall (f g : A -> A),
      (forall a, g (f a) = a) ->
      (forall a, f (g a) = a) ->
      (forall a b, f a = f b -> a = b) ->
      (forall a b, g a = g b -> a = b) ->
      Perm A.

Definition permt (A : AtomT) : Set := Perm A.


(**
  Application and inverses are straightforward to define.
*)
Definition perma (A : AtomT) (p : permt A) (a : A) :=
  let (f, _, _, _, _, _) := p in f a.

Definition pinv (A : AtomT) (p : permt A) :=
  let (f, g, p1, p2, p3, p4) := p in
    Perm_intro A g f p2 p1 p4 p3.

Implicit Arguments perma [A].
Implicit Arguments pinv [A].

Notation "p @ a" :=
  (perma p a)
  (at level 65, right associativity).

Notation "- p" :=
  (pinv p)
  (at level 35, right associativity).

Lemma perma_inverse :
  forall (A : AtomT) (p : permt A) (a : A),
  (- p) @ p @ a = a.
Proof.
  induction p; auto.
Qed.

Lemma perm_apply_inv :
  forall (A : AtomT) (p q : permt A),
  (forall a, p @ a = q @ a) ->
  (forall a, (- p) @ a = (- q) @ a).
Proof.
  intro n; induction p; induction q; unfold perma; simpl.
  intros H b.
  assert (J : f = f0); [apply fun_extensionality; auto | subst].
  assert (J : f0 (g b) = b); [auto | idtac].
  assert (L : f0 (g b) = f0 (g0 b)); [rewrite J; auto | auto].
Qed.

Lemma perma_eq :
  forall (A : AtomT) (p q : permt A),
  (forall a, p @ a = q @ a) -> p = q.
Proof.
  intros n p q H.
  assert (J : forall a, (- p) @ a = (- q) @ a).
  apply perm_apply_inv; auto.

  induction p; induction q; unfold perma in *;
    unfold pinv in *; simpl in *.

  assert (f = f0); [apply fun_extensionality; auto | subst].
  assert (g = g0); [apply fun_extensionality; auto | subst].
  assert (e = e3); [apply Proof_irrelevance; auto | subst].
  assert (e0 = e4); [apply Proof_irrelevance; auto | subst].
  assert (e1 = e5); [apply Proof_irrelevance; auto | subst].
  assert (e2 = e6); [apply Proof_irrelevance; auto | subst].
  reflexivity.
Qed.


(**
  The identity permutation is also fairly straightforward to define.
*)
Definition pid (A : AtomT) :=
  Perm_intro A (fun a => a) (fun a => a)
    (fun a => refl_equal a)
    (fun a => refl_equal a)
    (fun a b H => H)
    (fun a b H => H).

Implicit Arguments pid [A].

Lemma perma_id :
  forall (A : AtomT) (a : A),
  pid @ a = a.
Proof.
  auto.
Qed.


(**
  We need to do some work to show that the (obvious) composition
  of two permutations, in our representation, has the required properties.
*)
Definition pcompose (A : AtomT) (p q : permt A) : permt A.
  intros A.
  induction 1 as [f1 g1 p1 p2 p3 p4].
  induction 1 as [f2 g2 q1 q2 q3 q4].
  refine (Perm_intro A
    (fun x => f1 (f2 x))
    (fun x => g2 (g1 x))
    _ _ _ _).
  intros x; rewrite p1; rewrite q1; trivial.
  intros x; rewrite q2; rewrite p2; trivial.
  intros a b; eauto.
  intros a b; eauto.
Defined.

Implicit Arguments pcompose [A].

Notation "p :++ q" :=
  (pcompose p q)
  (at level 60, right associativity).

Lemma perma_compose :
  forall (A : AtomT) (p q : permt A) (a : A),
  (p :++ q) @ a = p @ q @ a.
Proof.
  induction p; induction q; auto.
Qed.


(**
  We need to do some work to show that the (obvious) way of extending
  permutations in our representation has the required properties.
*)

Definition pextend (A : AtomT) (ab : (A * A)) (p : permt A) : permt A.
  intros A ab; induction 1 as [f g p1 p2 p3 p4 p5 p6].
  refine (Perm_intro A
    (fun x => swapa A ab (f x))
    (fun x => g (swapa A ab x))
    _ _ _ _).
  intros x; autorewrite with nominal; trivial.
  intros x; rewrite p2; autorewrite with nominal; trivial.
  intros a b; eauto with nominal.
  intros a b H.
    assert (Q := p4 _ _ H).
    eauto with nominal.
Defined.

Implicit Arguments pextend [A].

Notation "x :+ p" :=
  (pextend x p)
  (at level 37, right associativity).

Lemma perma_extend :
  forall (A : AtomT) (p : permt A) (ab : A * A) (c : A),
  (ab :+ p) @ c = swapa A ab (p @ c).
Proof.
  induction p; auto.
Qed.

(* end hide *)

End PermutationsImplementation.

Export PermutationsImplementation.


(* ********************************************************************** *)
(** ** Theorems about atom permutations *)

Section PermutationEqualities.
Variable A : AtomT.

Lemma perma_double_inv_eq : forall (p : permt A), - - p = p.
Proof.
  intros p; apply perma_eq; intros a.
  replace a with (-p @ p @ a); auto with nominal.
  autorewrite with nominal.
  reflexivity.
Qed.

Lemma perma_compose_inv_eq : forall (p : permt A), -p :++ p = [].
Proof.
  intros p; apply perma_eq; intros a.
  autorewrite with nominal.
  reflexivity.
Qed.

Lemma perma_compose_inv_eq' : forall (p : permt A), p :++ -p = [].
Proof.
  intros p.
  pattern p at 1; replace p with (- - p).
  apply perma_compose_inv_eq.
  apply perma_double_inv_eq.
Qed.

Lemma perma_inverse' : forall (p : permt A) a, p @ -p @ a = a.
Proof.
  intros p a; pattern a at 2; rewrite <- perma_id.
  rewrite <- perma_compose.
  rewrite perma_compose_inv_eq'.
  reflexivity.
Qed.

Lemma perma_inv_id : forall (a : A), (- []) @ a = a.
Proof.
  intros a.
  rewrite <- (perma_id _ ((- []) @ a)).
  rewrite perma_inverse'.
  reflexivity.
Qed.

Lemma perma_inv_id_eq : - [] = ([] : permt A).
Proof.
  apply perma_eq.
  intros a; rewrite perma_id.
  apply perma_inv_id.
Qed.

Lemma perma_injective :
  forall (p : permt A) (a b : A),
  p @ a = p @ b -> a = b.
Proof.
  intros p a b H.
  assert (J : -p @ p @ a = -p @ p @ b); f_equal; trivial.
  autorewrite with nominal in J.
  assumption.
Qed.

Lemma perma_inv_compose :
  forall (p q : permt A) a, - (p :++ q) @ a = (-q) :++ (-p) @ a.
Proof.
  intros p q a; apply (perma_injective (p :++ q)).
  autorewrite with nominal.
  repeat rewrite perma_inverse'.
  rewrite <- (perma_compose _ p q).
  rewrite perma_inverse'.
  reflexivity.
Qed.

Lemma perma_inv_compose_eq :
  forall (p q : permt A), - (p :++ q) = (-q) :++ (-p).
Proof.
  intros p q; apply perma_eq; intros a; apply perma_inv_compose.
Qed.

Lemma perma_swap_commute :
  forall (p : permt A) (a b : A),
  p :++ [(a,b)] = [(p @ a, p @ b)] :++ p.
Proof.
  intros p a b; apply perma_eq; intros c.
  autorewrite with nominal.
  swapa_case ac bc a b c; try reflexivity.
  swapa_case ac' bc' (p @ a) (p @ b) (p @ c); try reflexivity.
  assert (a = c); eauto using perma_injective; contradiction.
  assert (b = c); eauto using perma_injective; contradiction.
Qed.

Lemma perma_swap_same : forall (a b : A), [(a, a)] @ b = b.
Proof.
  intros; autorewrite with nominal; reflexivity.
Qed.

Lemma perma_swap_same_eq : forall (a b : A), [(a, a)] = [].
Proof.
  intros a b; apply perma_eq; intros c; autorewrite with nominal.
  reflexivity.
Qed.

Lemma perma_swap_invol :
  forall (ab : A * A) (c : A), [ab] @ [ab] @ c = c.
Proof.
  intros ab c.
  induction ab.
  autorewrite with nominal.
  reflexivity.
Qed.

Lemma perma_swap_invol_eq :
  forall (ab : A * A), [ab] :++ [ab] = [].
Proof.
  intros ab; apply perma_eq.
  intros a; autorewrite with nominal.
  reflexivity.
Qed.

Lemma perma_swap_switch :
  forall (a b c : A), [(a, b)] @ c = [(b, a)] @ c.
Proof.
  intros a b c; autorewrite with nominal.
  swapa_case H J a b c; reflexivity.
Qed.

Lemma perma_swap_switch_eq :
  forall (a b : A), [(a, b)] = [(b, a)].
Proof.
  intros a b; apply perma_eq.
  apply perma_swap_switch.
Qed.

Lemma perma_inv_swap :
  forall (ab : A * A) (c : A), (- [ab]) @ c = [ab] @ c.
Proof.
  intros [a b] c.
  apply (perma_injective [(a, b)]).
  rewrite perma_swap_invol.
  rewrite perma_inverse'.
  reflexivity.
Qed.

Lemma perma_inv_swap_eq :
  forall (ab : A * A), - [ab] = [ab].
Proof.
  intros [a b]; apply perma_eq; intros c.
  apply perma_inv_swap.
Qed.

Lemma perma_swap_distrib :
  forall ab (c d x : A),
  [ab] @ [(c,d)] @ x =
  [(swapa A ab c, swapa A ab d)] @ [ab] @ x.
Proof.
  intros [a b] c d x.
  autorewrite with nominal.
  swapa_case cx dx c d x;
  swapa_case ad bd a b d;
  swapa_case ac bc a b c;
  swapa_case ax bx a b x;
  try reflexivity.
Qed.

End PermutationEqualities.

Hint Resolve perma_double_inv_eq : nominal.
Hint Rewrite perma_double_inv_eq : nominal.
Hint Resolve perma_compose_inv_eq perma_compose_inv_eq': nominal.
Hint Rewrite perma_compose_inv_eq perma_compose_inv_eq': nominal.
Hint Resolve perma_inverse' : nominal.
Hint Rewrite perma_inverse' : nominal.
Hint Resolve perma_inv_id perma_inv_id_eq : nominal.
Hint Rewrite perma_inv_id perma_inv_id_eq : nominal.
Hint Immediate perma_injective : nominal.
Hint Resolve perma_inv_compose perma_inv_compose_eq : nominal.
Hint Rewrite perma_inv_compose perma_inv_compose_eq : nominal.
Hint Resolve perma_swap_commute : nominal.
Hint Resolve perma_swap_same perma_swap_same_eq : nominal.
Hint Rewrite perma_swap_same perma_swap_same_eq : nominal.
Hint Resolve perma_swap_invol perma_swap_invol_eq : nominal.
Hint Rewrite perma_swap_invol perma_swap_invol_eq : nominal.
Hint Resolve perma_swap_switch perma_swap_switch_eq : nominal.
Hint Resolve perma_inv_swap perma_inv_swap_eq : nominal.
Hint Rewrite perma_inv_swap perma_inv_swap_eq : nominal.
Hint Resolve perma_swap_distrib : nominal.
