Require Import STLC.
Require Import Axioms.

Module UseSTLC (STLCImpl : STLC).
Import STLCImpl.

(*****************************************************************************)
(** Prove a specialized induction principle for terms *)

Theorem tm_induction :
  forall (P : tm -> Type) (F : aset tmvar),
  (P dot) ->
  (forall a, P (var a)) ->
  (forall s, P s -> forall t, P t -> P (app s t)) ->
  (forall a, ~ In a F -> forall T s, P s -> P (lam a T s)) ->
  forall x : tm, P x.
Proof.
  intros P F f_dot f_var f_app f_lam x.
  cut (forall p, P (perm tmP p x)).
  intros Main; assert (Q := Main []).
  autorewrite with nominal in Q; assumption.
  pattern x; apply tm_induction_weak.
    intros p; autorewrite with stlc; assumption.
    intros a p; autorewrite with stlc; apply f_var.
    intros s IHs t IHt p; autorewrite with stlc.
      apply f_app; trivial.
    intros a T s IHs p; autorewrite with stlc.
      set (a' := perm tmvarP p a) in *.
      set (s' := perm tmP p s) in *.
      case (atom_infinite tmvar (add a' (union (fvar s') F))); intros b B.
      rewrite (eq_lam a' b); auto with sets.
      apply f_lam; auto with sets.
        unfold s' in *.
        rewrite <- perm_compose.
        apply IHs.
Qed.

(**
  We can also show, trivially, that [tm_induction_weak] is derivable
  from [tm_induction].
*)
Theorem tm_induction_weak_derived :
  forall (P : tm -> Type),
  (P dot) ->
  (forall a, P (var a)) ->
  (forall s, P s -> forall t, P t -> P (app s t)) ->
  (forall a T s, P s -> P (lam a T s)) ->
  forall x : tm, P x.
Proof.
  intros P f_dot f_var f_app f_lam.
  intros x; pattern x.
  apply tm_induction with (F := @empty _ (asetR tmvar)); auto.
Qed.

(*****************************************************************************)
(** Support *)

Theorem supports_tm : forall t : tm, supports tmP (fvar t) t.
Proof.
  intros t; pattern t; apply tm_induction_weak; unfold supports; clear t.
  intros c d C D; autorewrite with stlc; reflexivity.
  intros x c d C D.
    autorewrite with stlc in C.
    autorewrite with stlc in D.
    destruct_neg_singleton C.
    destruct_neg_singleton D.
    autorewrite with stlc.
    unfold tmvarP; rewrite perm_atom.
    autorewrite with nominal.
    reflexivity.
  intros s IHs t IHt c d C D.
    autorewrite with stlc.
    autorewrite with stlc in C.
    autorewrite with stlc in D.
    rewrite IHs; auto with sets.
    rewrite IHt; auto with sets.
  intros x T s IHs c d C D.
    autorewrite with stlc.
    autorewrite with stlc in C.
    autorewrite with stlc in D.
    assert (C' := remove_neg_elim_strong_1 _ _ _ C); clear C.
    assert (D' := remove_neg_elim_strong_1 _ _ _ D); clear D.
    inversion_clear C' as [C | [C C'']]; inversion_clear D' as [D | [D D'']].
      subst.
        autorewrite with nominal.
        reflexivity.
      subst.
        unfold tmvarP; rewrite perm_atom; autorewrite with nominal.
        symmetry.
        apply eq_lam.
        assumption.
      subst.
        unfold tmvarP; rewrite perm_atom; autorewrite with nominal.
        symmetry.
        rewrite perm_swap_switch.
        apply eq_lam.
        assumption.
      unfold tmvarP; rewrite perm_atom; autorewrite with nominal.
        rewrite IHs; trivial.
Qed.

Hint Resolve supports_tm : stlc.

(*****************************************************************************)
(** Define substitution *)

Section Substitution.
Variable y : tmvar.
Variable s : tm.

Definition subst : tm -> tm :=
  tm_rec tm
    dot
    (fun x => if atom_eqdec tmvar x y then s else (var x))
    (fun s s' t t' => app s' t')
    (fun x T t t' => lam x T t')
    (add y (fvar s)).

Lemma subst_supp_dot :
  supports tmP (add y (fvar s)) dot.
Proof.
  unfold supports; intros c d C D.
  autorewrite with stlc.
  reflexivity.
Qed.

Lemma subst_supp_var :
  supports (tmvarP ^-> tmP) (add y (fvar s))
    (fun x : tmvar => if atom_eqdec tmvar x y then s else var x).
Proof.
  apply supports_fun; intros c d x C D.
  case (atom_eqdec tmvar x y); intros eq.
    destruct_neg_add C H J.
    destruct_neg_add D K L.
    case (atom_eqdec tmvar (perm tmvarP [(c,d)] x) y).
      intros M.
      rewrite swap_non_supports with (F := fvar s); auto with stlc.
      intros M.
      subst.
      unfold tmvarP in M; rewrite perm_atom in M.
      autorewrite with nominal in M.
      intuition.
    destruct_neg_add C H J.
    destruct_neg_add D K L.
    case (atom_eqdec tmvar (perm tmvarP [(c,d)] x) y).
      intros M.
      assert (N := perm_swap_move' _ _ _ _ _ _ M).
      unfold tmvarP in N; rewrite perm_atom in N.
      autorewrite with nominal in N.
      contradiction.
      intros M.
      autorewrite with stlc.
      reflexivity.
Qed.

Lemma subst_supp_app :
  supports (tmP ^-> tmP ^-> tmP ^-> tmP ^-> tmP)
    (add y (fvar s)) (fun _ s' _ t' : tm => app s' t').
Proof.
  unfold supports; intros c d C D.
  apply fun_extensionality; intros q.
  apply fun_extensionality; intros q'.
  apply fun_extensionality; intros r.
  apply fun_extensionality; intros r'.
  repeat rewrite perm_fun.
  autorewrite with nominal.
  apply (perm_injective _ _ tmP ([(c, d)])).
  autorewrite with nominal stlc.
  reflexivity.
Qed.

Lemma subst_supp_lam :
  supports (tmvarP ^-> tyP ^-> tmP ^-> tmP ^-> tmP)
    (add y (fvar s))
    (fun (x : tmvar) (T : ty) (_ t' : tm) => lam x T t').
Proof.
  unfold supports; intros c d C D.
  apply fun_extensionality; intros x.
  apply fun_extensionality; intros T.
  apply fun_extensionality; intros t.
  apply fun_extensionality; intros t'.
  repeat rewrite perm_fun.
  autorewrite with nominal.
  apply (perm_injective _ _ tmP [(c,d)]).
  unfold tyP.
  autorewrite with nominal stlc.
  simpl.
  reflexivity.
Qed.

Hint Resolve subst_supp_dot subst_supp_var subst_supp_app subst_supp_lam : stlc.

Theorem subst_dot : subst dot = dot.
Proof.
  unfold subst.
  autorewrite with stlc.
  reflexivity.
Qed.

Theorem subst_var_eq : subst (var y) = s.
Proof.
  unfold subst.
  autorewrite with stlc nominal.
  reflexivity.
Qed.

Theorem subst_var_neq : forall x, x <> y -> subst (var x) = (var x).
Proof.
  unfold subst.
  intros x H; autorewrite with stlc nominal.
  reflexivity.
Qed.

Theorem subst_app :
  forall q r, subst (app q r) = app (subst q) (subst r).
Proof.
  unfold subst.
  intros q r; autorewrite with stlc.
  reflexivity.
Qed.

Theorem subst_lam :
  forall x T t, x <> y -> ~ In x (fvar s) ->
  subst (lam x T t) = lam x T (subst t).
Proof.
  unfold subst.
  intros x T t H J.
  rewrite tm_rec_lam with (RP := tmP); try reflexivity; auto with stlc sets.
  exists x; split; auto with sets.
  intros x' T' t'.
  unfold freshP.
  exists (fvar (lam x x' t')); split; auto with stlc.
  autorewrite with stlc; auto with sets.
Qed.

Theorem supports_subst :
  supports (tmP ^-> tmP) (add y (fvar s)) subst.
Proof.
  unfold subst.
  apply tm_rec_supp; auto with stlc.
  case (atom_infinite tmvar (add y (fvar s))); intros x X.
  exists x; split; auto with sets.
  intros x' T' t'.
  unfold freshP.
  exists (fvar (lam x x' t')); split; auto with stlc.
  autorewrite with stlc; auto with sets.
Qed.

End Substitution.

Hint Resolve subst_dot subst_var_eq subst_var_neq subst_app subst_lam : stlc.
Hint Rewrite subst_dot subst_var_eq subst_app : stlc.
Hint Rewrite subst_var_neq using congruence : stlc.
Hint Rewrite subst_lam using (auto with stlc nominal; fail) : stlc.

(*****************************************************************************)
(** ** Simple substitution lemmas *)

Notation "'existsType' x : t , p" :=
  (sigT (fun x : t => p))
  (at level 200, x ident, format "'existsType' '/ ' x : t , '/ ' p") :
  type_scope.

Lemma tm_induction' :
  forall (P : tm -> Type),
  (P dot) ->
  (forall (a : tmvar), P (var a)) ->
  (forall s : tm, P s-> forall t : tm, P t -> P (app s t)) ->
  (existsType E : aset tmvar,
    forall (a : tmvar), ~ In a E -> forall (T : ty) (s : tm),
      P s -> P (lam a T s)) ->
  forall x : tm, P x.
Proof.
  intros P H_dot H_var H_app H_lam.
  elim H_lam; intros E H_lam' x.
  apply (tm_induction P E);
  [exact H_dot | exact H_var | exact H_app | exact H_lam'].
Qed.

Notation "M ^[ x := N ]" := (subst x N M) (at level 59, left associativity).

(*****************************************************************************)

Lemma subst_var_same' :
  forall (x : tmvar) (e : tm) (y : tmvar),
  y = x -> (var y) ^[x := e] = e.
Proof.
  intros x e y H; rewrite H; apply subst_var_eq.
Qed.

Hint Rewrite subst_var_same' using congruence : stlc.

Ltac subst_var_simpl H x y :=
  match goal with
  | |- context [subst x ?e (var y)] =>
      (progress (repeat
        (rewrite (subst_var_same' x e y); [idtac | congruence]))) ||
      (progress (repeat
        (rewrite (subst_var_neq x e y); [idtac | congruence]))) ||
      (case (atom_eqdec _ x y); intro H;
        [repeat (rewrite (subst_var_same' x e y)); [idtac | congruence] |
         repeat (rewrite (subst_var_neq x e y)); [idtac | congruence]])
  end.

Ltac stlc_simpl :=
  repeat (
    progress (
      unfold tmvarP ||
      rewrite perm_atom ||
      autorewrite with stlc nominal ||
      swapa_simpl)).

Lemma fvar_swap :
  forall x y z t,
  z <> x -> z <> y ->
  ~ In z (fvar t) ->
  ~ In z (fvar (perm tmP [(x, y)] t)).
Proof.
  intros x y z t zx zy.
  pattern t; apply tm_induction'.
    (* unit *)
    stlc_simpl; trivial.
    (* var *)
    intros w H.
    rewrite fvar_var in H.
    stlc_simpl.
    swapa_case xw yw x y w; auto with sets.
    (* app *)
    intros u IHu v IHv H.
    rewrite fvar_app in H.
    stlc_simpl.
    destruct_neg_union H J K.
    auto with sets.
    (* var *)
    exists (add z (fvar t)).
    intros w w_fresh T u IHu H.
    destruct_neg_add w_fresh wz wt.
    rewrite fvar_lam in H.
    destruct_neg_remove H.
      (* z = w *)
      congruence.
      (* z not in (fvar s) *)
      stlc_simpl.
      swapa_case xw yw x y w; auto with sets.
Qed.

Hint Resolve fvar_swap : stlc.

Lemma subst_equivariant :
  forall z v x y t,
  perm tmP [(x, y)] (t ^[z := v]) =
    (perm tmP [(x, y)] t) ^[(perm tmvarP [(x, y)] z) := (perm tmP [(x, y)] v)].
Proof.
  intros z v x y t.
  pattern t; apply tm_induction'.
    (* unit *)
    stlc_simpl; trivial.

    (* var *)
    intros a.
    subst_var_simpl za z a; stlc_simpl; trivial.
    swapa_case xa ya x y a;
    swapa_case xz yz x y z; stlc_simpl; trivial.

    (* app *)
    intros u1 IHu1 u2 IHu2.
    rewrite subst_app.
    rewrite perm_app.
    rewrite IHu1.
    rewrite IHu2.
    stlc_simpl; trivial.

    (* lam *)
    exists (add x (add y (add z (fvar v)))).
    intros a Ha T s IHs.
    destruct_neg_add Ha ax Ha1.
    destruct_neg_add Ha1 ay Ha2.
    destruct_neg_add Ha2 az av.
    stlc_simpl.
    rewrite IHs.
    unfold tmvarP;
    swapa_case xz yz x y z;
    stlc_simpl; trivial.
Qed.

Hint Rewrite subst_equivariant : stlc.

Lemma fvar_subst :
  forall (x : tmvar) (M N : tm),
  Subset (fvar (M ^[x := N])) (union (remove x (fvar M)) (fvar N)).
Proof.
  intros x M N.
  pattern M; apply tm_induction'.

    (* Case: unit *)
    autorewrite with stlc; auto with sets.

    (* Case: var *)
    intros y.
    subst_var_simpl xy x y.
      (* x = y *)
      auto with sets.
      (* x <> y *)
      autorewrite with stlc.
      unfold Subset; intros z H.
      destruct_singleton H; rewrite H; auto with sets.

    (* Case: app *)
    intros s IHs t IHt.
    autorewrite with stlc.
    autorewrite with sets.
    rewrite union_union_distrib.
    auto with sets.

    (* Case: lam *)
    exists (add x (fvar N)).
    intros y y_fresh T s IHs.
    destruct_neg_add y_fresh yx y_fresh'.
    autorewrite with stlc.
    unfold Subset; intros z H.
    destruct_remove H J K.
    assert (L := IHs z K).
    destruct_union L.
    destruct_remove L L1 L2; auto with sets.
    auto with sets.
Qed.

Lemma subst_not_fv :
  forall (x : tmvar) (N M: tm),
  ~ In x (fvar M) ->
  M ^[x := N] = M.
Proof.
  intros x N M.
  pattern M; apply tm_induction'.

    (* unit *)
    autorewrite with stlc; trivial.

    (* var *)
    intros a H.
    rewrite fvar_var in H.
    destruct_neg_singleton H.
    autorewrite with stlc; trivial.

    (* app *)
    intros s IHs t IHt H.
    rewrite fvar_app in H.
    destruct_neg_union H Hs Ht.
    autorewrite with stlc.
    rewrite IHs; trivial.
    rewrite IHt; trivial.

    (* lam *)
    exists (add x (fvar N)).
    intros y y_fresh T s IHs H.
    destruct_neg_add y_fresh yx yN.
    autorewrite with stlc.
    rewrite IHs; trivial.
    rewrite fvar_lam in H.
    destruct_neg_remove H; [congruence | trivial].
Qed.

Hint Rewrite subst_not_fv using solve [auto with stlc] : stlc.

(* This is a useful tactic (I think), but not needed in anywhere here. *)
Ltac fresh_atom name set hyp_name :=
  let H := fresh in (
    assert (H := atom_infinite _ set);
    elim H; clear H;
    intros name hyp_name
  ).

(* This lemma isn't used later.  It's just here as an example. *)
Lemma subst_commute :
  forall (x y : tmvar) (M N L : tm),
  x <> y ->
  ~ In x (fvar L) ->
  M ^[x := N] ^[y := L] = M ^[y := L] ^[x := N ^[y := L]].
Proof.
  intros x y M N L xy xL.
  pattern M; apply tm_induction'.

    (* dot *)
    autorewrite with stlc; trivial.

    (* var *)
    intro z.
    subst_var_simpl xz x z;
    subst_var_simpl yz y z;
    autorewrite with stlc; trivial.

    (* app *)
    intros s IHs t IHt.
    autorewrite with stlc.
    rewrite IHs; trivial.
    rewrite IHt; trivial.

    (* lam *)
    exists (union (add x (fvar N)) (add y (fvar L))).
    intros z z_fresh T s IHs.
    destruct_neg_union z_fresh z_fresh1 z_fresh2.
    destruct_neg_add z_fresh1 zx zN.
    destruct_neg_add z_fresh2 zy zL.
    autorewrite with stlc.
    rewrite IHs; trivial.
    rewrite subst_lam; trivial.
    intro J.
    assert (J' := fvar_subst _ _ _ _ J).
    destruct_union J'.
    destruct_remove J' J1 J2; contradiction.
    contradiction.
Qed.

(* This lemma isn't used later.  It's just here as an example. *)
Lemma subst_fresh_var :
  forall (x y : tmvar) (M : tm),
  ~ In y (fvar M) ->
  M ^[x := var y] = perm tmP [(x, y)] M.
Proof.
  intros x y M.
  pattern M; apply tm_induction'.

    (* unit *)
    autorewrite with stlc; trivial.

    (* var *)
    intros z y_fresh.
    rewrite fvar_var in y_fresh.
    destruct_neg_singleton y_fresh.
    subst_var_simpl xz x z; stlc_simpl; trivial.

    (* app *)
    intros s IHs t IHt y_fresh.
    rewrite fvar_app in y_fresh.
    destruct_neg_union y_fresh ys yt.
    autorewrite with stlc.
    rewrite IHs; trivial; rewrite IHt; trivial.

    (* lam *)
    exists (add x (fvar (var y))).
    intros z z_fresh T s IHs y_fresh.
    destruct_neg_add z_fresh zx z_var_y.
    rewrite fvar_lam in y_fresh.
    destruct_neg_remove y_fresh.
      (* y = z *)
      absurd (z <> x); trivial.
      rewrite fvar_var in z_var_y.
      destruct_neg_singleton z_var_y.
      congruence.
      (* y not in (fvar s) *)
      autorewrite with stlc.
      rewrite IHs; trivial.
      rewrite fvar_var in z_var_y.
      destruct_neg_singleton z_var_y.
      stlc_simpl; trivial.
Qed.

(* This is the critical lemma for detminacy of the evaluation relation. *)
Lemma subst_equal :
  forall (x y : tmvar) (M N : tm),
  ~ In y (fvar M) ->
  M ^[x := N] = (perm tmP [(y, x)] M) ^[y := N].
Proof.
  intros x y M N.
  pattern M; apply tm_induction'.

    (* unit *)
    autorewrite with stlc; trivial.

    (* var *)
    intros z y_fresh.
    rewrite fvar_var in y_fresh.
    destruct_neg_singleton y_fresh.
    subst_var_simpl xz x z; stlc_simpl; trivial.

    (* app *)
    intros s IHs t IHt y_fresh.
    rewrite fvar_app in y_fresh.
    destruct_neg_union y_fresh ys yt.
    autorewrite with stlc.
    rewrite IHs; trivial; rewrite IHt; trivial.

    (* lam *)
    exists (add x (add y (fvar N))).
    (* NB: The atom "y" is added above solely for the purpose of creating a
       contradiction below (in the case "y = z").  I think this trick is one
       of the most noteworthy technical aspects of the development, since this
       is perhaps the only case where I felt like I was doing something that
       didn't resemble anything I would do on paper and I believe it will be
       necessary to use it often. *)
    intros z z_fresh T s IHs y_fresh.
    destruct_neg_add z_fresh zx z_fresh'.
    destruct_neg_add z_fresh' zy zN.
    rewrite fvar_lam in y_fresh.
    destruct_neg_remove y_fresh.
      (* y = z *)
      congruence.
      (* y not in (fvar s) *)
      stlc_simpl.
      rewrite IHs; trivial.
Qed.


(*****************************************************************************)
(** CBV Reduction *)

Inductive value : tm -> Prop :=
  | unit_value :
      value dot
  | abs_value :
      forall (x : tmvar) (T : ty) (t : tm),
      value (lam x T t).

Inductive cbv : tm -> tm -> Prop :=
  | cbv_left :
      forall (t t' u : tm),
      cbv t t' ->
      cbv (app t u) (app t' u)
  | cbv_right :
      forall (t u u' : tm),
      value t ->
      cbv u u' ->
      cbv (app t u) (app t u')
  | cbv_beta :
      forall (x : tmvar) (T : ty) (t u : tm),
      value u ->
      cbv (app (lam x T t) u) (t ^[x := u]).

Lemma values_are_normal_forms :
  forall (t t' : tm), value t -> ~ cbv t t'.
Proof.
  intros t t' H J.
  induction J; inversion H; stlc_discriminate.
Qed.

Lemma cbv_deterministic :
  forall (t t' t'' : tm),
  cbv t t' -> cbv t t'' -> t' = t''.
Proof.
  intros t t' t'' step1.
  generalize t''; clear t''.
  induction step1.
    (* cbv_left *)
    intros v J.
    inversion J; stlc_injection.
      (* cbv_left *)
      rewrite e in H0.
      assert (Q := IHstep1 t'0 H0); congruence.
      (* cbv_right *)
      rewrite e in H0.
      absurd (cbv t t'); try apply values_are_normal_forms; trivial.
      (* cbv_beta *)
      assert (Q : value t).
      rewrite <- e; apply abs_value.
      absurd (cbv t t'); try apply values_are_normal_forms; trivial.
    (* cbv_right *)
    intros v J.
    inversion J; stlc_injection.
      (* cbv_left *)
      rewrite e in H1.
      absurd (cbv t t'); try apply values_are_normal_forms; trivial.
      (* cbv_right *)
      rewrite e0 in H2.
      assert (Q := IHstep1 u'0 H2); congruence.
      (* cbv_beta *)
      rewrite e0 in H1.
      absurd (cbv u u'); try apply values_are_normal_forms; trivial.
    (* cbv_beta *)
    intros v J.
    inversion J; stlc_injection.
      (* cbv_left *)
      assert (Q : value t0).
      rewrite e; apply abs_value.
      absurd (cbv t0 t'); try apply values_are_normal_forms; trivial.
      (* cbv_right *)
      rewrite e0 in H2.
      absurd (cbv u u'); try apply values_are_normal_forms; trivial.
      assert (K := e); assert (L := e0); clear e e0 H0.
      stlc_injection.
      decompose [and or] o; clear o.
      congruence.
      (* cbv_beta *)
      rewrite L; rewrite H5.
      apply subst_equal; trivial.
Qed.

Lemma value_equivariant :
  forall x y t,
  value t -> value (perm tmP [(x, y)] t).
Proof.
  intros x y t H.
  induction H; autorewrite with stlc.
    (* unit_value *)
    apply unit_value.
    (* abs_value *)
    apply abs_value.
Qed.

Lemma cbv_equivariant :
  forall x y t t',
  cbv t t' -> cbv (perm tmP [(x, y)] t) (perm tmP [(x, y)] t').
Proof.
  intros x y t u step.
  induction step; autorewrite with stlc.
    (* cbv_left *)
    apply cbv_left; trivial.
    (* cbv_right *)
    apply cbv_right; trivial.
    apply value_equivariant; trivial.
    (* cbv_beta *)
    apply cbv_beta.
    apply value_equivariant; trivial.
Qed.


(*****************************************************************************)
(** Size function *)

Inductive size : tm -> nat -> Prop :=
  | size_unit : size dot 1
  | size_var : forall x, size (var x) 1
  | size_app :
      forall t u m n,
      size t m -> size u n -> size (app t u) (m + n + 1)
  | size_lam :
      forall x T t m,
      size t m -> size (lam x T t) (m + 1).

Lemma size_total :
  forall t, {n : nat | size t n}.
Proof.
  intros t; pattern t; apply tm_induction'.
    (* unit *)
    exists 1; apply size_unit.
    (* var *)
    intro x; exists 1; apply size_var.
    (* app *)
    intros u IHu v IHv.
    inversion_clear IHu as [nu IHu'].
    inversion_clear IHv as [nv IHv'].
    exists (nu + nv + 1).
    apply size_app; trivial.
    (* lam *)
    exists (empty (e := asetR tmvar)).
    intros x x_fresh T u IHu.
    inversion_clear IHu as [nu IHu'].
    exists (nu + 1).
    apply size_lam; trivial.
Defined.

Lemma size2 : forall x T, proj1_sig (size_total (lam x T (var x))) = 2.
Admitted. (* XXX size2 (an example lemma) *)

(*****************************************************************************)
(** Compiling to SK-combinators *)

Inductive SKtm : Set :=
  | SKvar : tmvar -> SKtm
  | S : SKtm
  | K : SKtm
  | SKapp : SKtm -> SKtm -> SKtm.

Fixpoint SKperm (p : permt tmvar) (s : SKtm) {struct s} : SKtm :=
  match s with
  | SKvar z => SKvar (perm tmvarP p z)
  | S => S
  | K => K
  | SKapp q r => SKapp (SKperm p q) (SKperm p r)
  end.

Definition SKP : PsetT tmvar SKtm.
  refine (mkPset SKperm _ _).
  induction x; simpl; autorewrite with nominal; congruence.
  induction x; simpl; autorewrite with nominal; congruence.
Defined.

Fixpoint SKfvar (s : SKtm) : aset tmvar :=
  match s with
  | SKvar x => singleton x
  | S => empty
  | K => empty
  | SKapp q r => union (SKfvar q) (SKfvar r)
  end.

Inductive SKabs : tmvar -> SKtm -> SKtm -> Prop :=
  | SKabs_I :
      forall x,
      SKabs x (SKvar x) (SKapp (SKapp S K) K)
  | SKabs_K :
      forall x s,
      ~ In x (SKfvar s) ->
      SKabs x s (SKapp K s)
  | SKabs_S :
      forall x r s r' s',
      SKabs x r r' ->
      SKabs x s s' ->
      SKabs x (SKapp r s) (SKapp (SKapp S r') s').

Inductive SKcomp : tm -> SKtm -> Prop :=
  | SKcomp_var :
      forall x,
      SKcomp (var x) (SKvar x)
  | SKcomp_app :
      forall t u q r,
      SKcomp t q ->
      SKcomp u r ->
      SKcomp (app t u) (SKapp q r)
  | SKcomp_lam :
      forall x T t s s',
      SKcomp t s ->
      SKabs x s s' ->
      SKcomp (lam x T t) s'.

Lemma SKvar_equivariant :
  forall p z,
  perm SKP p (SKvar z) = SKvar (perm tmvarP p z).
Proof. trivial. Qed.
Hint Rewrite SKvar_equivariant : stlc.

Lemma S_equivariant :
  forall p,
  perm SKP p S = S.
Proof. trivial. Qed.
Hint Rewrite S_equivariant : stlc.

Lemma K_equivariant :
  forall p,
  perm SKP p K = K.
Proof. trivial. Qed.
Hint Rewrite K_equivariant : stlc.

Lemma SKapp_equivariant :
  forall p r s,
  perm SKP p (SKapp r s) =
    SKapp (perm SKP p r) (perm SKP p s).
Proof. trivial. Qed.
Hint Rewrite SKapp_equivariant : stlc.

Lemma SKfvar_equivariant :
  forall p z s,
  In z (SKfvar s) ->
  In (perm tmvarP p z) (SKfvar (perm SKP p s)).
Proof.
  intros p z s H.
  induction s.
    (* SKvar *)
    stlc_simpl.
    simpl in H; destruct_singleton H.
    subst; simpl; auto with sets.
    (* S *)
    unfold SKfvar in H.
    destruct_empty.
    (* K *)
    unfold SKfvar in H.
    destruct_empty.
    (* SKapp *)
    stlc_simpl.
    simpl SKfvar.
    simpl in H; destruct_union H; auto with sets.
Qed.

Lemma SKabs_equivariant :
  forall p z s s',
  SKabs z s s' ->
  SKabs (perm tmvarP p z) (perm SKP p s) (perm SKP p s').
Proof.
  intros p z s s' H; induction H; stlc_simpl.
  (* SKabs_I *)
  apply SKabs_I.
  (* SKabs_K *)
  apply SKabs_K.
  intro J; apply H.
  assert (K := SKfvar_equivariant (-p) _ _ J).
  rewrite <- perm_atom in K.
  autorewrite with nominal in K.
  trivial.
  (* SKabs_S *)
  apply SKabs_S; trivial.
Qed.

Lemma SKcomp_equivariant :
  forall p t s,
  SKcomp t s ->
  SKcomp (perm tmP p t) (perm SKP p s).
Proof.
  intros p t s H; induction H.
    (* SKcomp_var *)
    stlc_simpl; apply SKcomp_var.
    (* SKcomp_app *)
    stlc_simpl; apply SKcomp_app; trivial.
    (* SKcomp_lam *)
    stlc_simpl.
    eapply SKcomp_lam.
    apply IHSKcomp.
    rewrite <- perm_atom.
    apply SKabs_equivariant; trivial.
Qed.

End UseSTLC.
